
#include <stdlib.h>
#include <string.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>

#include "PortDef.h"
#include "LcdCtrl.h"

#define SELECT_L	{CS1_H;CS2_L;}
#define SELECT_R	{CS1_L;CS2_H;}

#define PALETTETOP	5 // pbgf[^[擪ATCYQoCg
#define FONT_OFFSET 6 // tHgԃf[^[擪

void InitPort(){
	uint8_t musk;
	musk = ((1<<BIT_RS)|(1<<BIT_RW)|(1<<BIT_E)|(1<<BIT_CS1)|(1<<BIT_CS2)|(1<<BIT_RST));
	PORT_CTRL &= ~(musk);
	DDR_CTRL |= musk;
	PORT_DATA = 0;
	DDR_DATA = 0;
}

// kbcrW[rbg҂
void BusyCheck(){
	DDR_DATA = 0;
	PORT_DATA = 0xFF; // vAbvOFF;
	RS_L;
	RW_H;
	E_H;
#if !IODEBUG
	_delay_us(1);
	while(PIN_DATA & 0x80);
#endif	
	E_L;
	RW_L;
}

// kbcR}hǂݍ
uint8_t ReadLcdCmd(){
	uint8_t ret;
	E_L;
	DDR_DATA = 0x00;
	PORT_DATA = 0xFF; 
	RS_L;
	RW_H;
	E_H;
	_delay_us(1); // read delay
	ret = PIN_DATA;
	E_L;
	RW_L;
	return ret;
}

// kbcf[^[ǂݍ
uint8_t ReadLcdData(){
	BusyCheck();
	uint8_t ret;
	DDR_DATA = 0x00;
	PORT_DATA = 0xFF; // Pull up Off
	RS_H;
	RW_H;
	E_H;
	_delay_us(1); // read delay
	ret = PIN_DATA;
	E_L;
	RW_L;
	RS_L;
	return ret;
}

// kbcR}h
void WriteLcdCmd(uint8_t ch){
	BusyCheck();
	DDR_DATA = 0xFF;
	RS_L;
	RW_L;
	PORT_DATA = ch;
	E_H;
	_delay_us(1);
	E_L;
	RS_L;
	DDR_DATA = 0;
}

// kbcf[^[
void WriteLcdData(uint8_t ch){
	BusyCheck();
	DDR_DATA = 0xFF;
	RS_H;
	RW_L;
	PORT_DATA = ch;
	E_H;
	_delay_us(1);
	E_L;
	RS_L;
	DDR_DATA = 0;
}

// kbc\Jn@on != 0@ŕ\ on == 0 Ŕ\
void DisplayOn(uint8_t on){
	if(on){
	 	SELECT_L
		WriteLcdCmd(0x3F);
	 	SELECT_R
		WriteLcdCmd(0x3F);
	}else{
		SELECT_L
		WriteLcdCmd(0x3E);
		SELECT_R
		WriteLcdCmd(0x3E);
	}
}

// ݃Cw肷 line = C
void SetStartLine(int line){
	uint8_t cmd;
	cmd = 0xC0;
	cmd |= (line & 0x3F);
	SELECT_L
	WriteLcdCmd(cmd);
	SELECT_R
	WriteLcdCmd(cmd);
}

// ݃y[Ww肷@page = y[W
void SetPage(int page){
	uint8_t cmd;
	cmd = 0xB8;
	cmd |= (page & 0x07);
	WriteLcdCmd(cmd);
}

// ݗiRjw肷B
void SetColumn(int column){
	uint8_t cmd;
	cmd = 0x40;
	cmd |= (column & 0x7F);
	WriteLcdCmd(cmd);
}

// kbc̃y[WRAMɓǂݍ
// ߂Fǂݍ񂾃oCg
int ReadLcdPage(int column,int page,uint8_t data[],int num){
	if(column < 64){
		SELECT_L
		SetColumn(column);
		SetPage(page);
		ReadLcdData();
		SELECT_R
		SetColumn(0);
		SetPage(page);
		ReadLcdData();
	}else{
		SELECT_R
		SetColumn(column-64);
		SetPage(page);
		ReadLcdData();
	}
	int i;
	for(i=0;i<num;i++){
		if((i+column)<64)SELECT_L
		else SELECT_R
		*data = ReadLcdData();
		data++;
	}
	return i;
}

// q`lkbcy[Wɏ
void WriteLcdPageRam(int column,int page,const uint8_t data[],int num){
	if(column < 64){
		SELECT_L;
		SetColumn(column);
		SetPage(page);
		SELECT_R;
		SetColumn(0);
		SetPage(page);
	}else{
		SELECT_R;
		SetColumn(column-64);
		SetPage(page);
	}
	int i;
	for(i=0;i<num;i++){
		if((i+column)<64){SELECT_L;}
		else {SELECT_R;}
		WriteLcdData(*data);
		data++;
	}
}

// vOkbcy[Wɏ
void WriteLcdPagePrgm(int column,int page,const prog_uchar dat[],int num){ 
// dat͒ډ摜̐擪w
	if(column < 64){
		SELECT_L;
		SetColumn(column);
		SetPage(page);
		SELECT_R;
		SetColumn(0);
		SetPage(page);
	}else{
		SELECT_R;
		SetColumn(column-64);
		SetPage(page);
	}
	for(int i=0;i<num;i++){
		if((i+column)<64){SELECT_L;}
		else{SELECT_R;}
		WriteLcdData(pgm_read_byte(&dat[i]));
	}
}
		
//kbcp^[Ŗ
void FillChipMemory(uint8_t pattern){
	uint8_t x,y;
	for(y=0;y<8;y++){
		SetPage(y);
		SetColumn(0);
		for(x=0;x<64;x++){
			WriteLcdData(pattern);
		}
	}
}

// kbc
void InitLcd(){
	InitPort();
	RST_L;
	_delay_ms(5);
	RST_H;
	_delay_ms(30);
	SetStartLine(0);
	SELECT_L

	while(!(ReadLcdCmd() & 0x20));
	FillChipMemory(0x55);
	SELECT_R
	FillChipMemory(0x55);
	DisplayOn(1);	
	while((ReadLcdCmd() & 0x20));
}

// kbcy[W͈̔͂q`lɓǂݍ
// ߂ɂP
int ReadPageRectRam(int left,int topPage
	,int right,int bottomPage,uint8_t data[]){
	for(int y = topPage; y < bottomPage;y++){
		ReadLcdPage(left,y,data,right-left+1);
		data += (right-left+1);
	}
	return 1;
}

// q`lkbcy[W͈̔͂ɏ 
// ߂ɂP
int WritePageRectRam(int left,int topPage
	,int right,int bottomPage,const uint8_t data[]){
	for(int y = topPage; y < bottomPage;y++){
		WriteLcdPageRam(left,y,data,right-left+1);
		data+=(right-left+1);
	}
	return 1;
}

// q`lkbcy[W͈̔͂ɏ ʒuwRECT
int WritePageRectRamR(RECT rect,const uint8_t data[]){
	return WritePageRectRam(rect.left,rect.top,rect.right,rect.bottom,data);
}


// vO璼LCDɔ͈͂o data͒ډ摜w
int WritePageRectPrgm(int left,int topPage
	,int right,int bottomPage,const prog_uchar data[]){
	int cnt = 0;
	int length = right-left+1;
	for(int y = topPage;y<bottomPage;y++){
		WriteLcdPagePrgm(left,y,&data[cnt],length);
		cnt += length;
	}
	return 1;
}


// ʃf[^[vOkbcɏ
// ʃf[^[͂kbcCAEgōwb_[t
// ߂FŔO
int WriteLcdFromPrgm(int left,int top,const prog_uchar src[]){
	const prog_uchar *pSrc = &src[0];
	int bpp = pgm_read_byte(pSrc);pSrc++;
	int pictOfs = PALETTETOP+(3*(1<<bpp))-1;
	int width = ((int)pgm_read_byte(pSrc)<<8);pSrc++;
	width |= pgm_read_byte(pSrc);pSrc++;
	int hight = ((int)pgm_read_byte(pSrc)<<8);pSrc++;
	hight |= pgm_read_byte(pSrc);src++;
	int right = left + width-1;
	int bottom = top + hight-1;
	int t = top/8;
	int b = bottom/8+1; 
	if((!(top%8))&&(!((bottom+1)%8))){ //㉺ƂɃoCgEȂ̂ŒP`
		WritePageRectPrgm(left,t,right,b,&src[pictOfs]);
		return 1; // Ń^[
	}
	int lShift = ((top)%8); // oCgÊ
	// sڂ̃}XN
	uint8_t topMusk = 0;
	if(lShift){
		for(int i=0;i<lShift;i++){
			topMusk = topMusk << 1;
			topMusk |= 1;
		}
	}// topMusk̓obN}XN

	// Ō̍s̃}XN

	int rShift = ((bottom+1)%8);
	uint8_t endMusk = 0;
	if(rShift){
		for(int i=0;i<rShift;i++){
			endMusk = endMusk << 1;
			endMusk |= 1;
		}
	}
	// endMusk̓f[^[}XN	

	uint8_t *lineBuff = malloc(width);
	uint16_t *shiftBuff = malloc(width<<1);
	uint8_t *pLine;
	uint16_t *pShift;
	pLine = lineBuff;
	if((b-t)<2){ // s̏̂
		if(rShift)topMusk = topMusk | ~endMusk;
		ReadLcdPage(left,t,pLine,width);
		for(int x= 0;x<width;x++){
			pLine[x] &= topMusk;
			pLine[x] |= ((pgm_read_byte(&src[x+pictOfs]) << lShift) & ~topMusk);
		}
		WriteLcdPageRam(left,t,pLine,width);
	}else{	// s
		for(int i=0;i<width;i++){// Vtgobt@NA
			shiftBuff[i]=0;
		}
		for(int y=0;y<b-t;y++){
			pShift = shiftBuff;
			pLine = lineBuff;
			ReadLcdPage(left,y+t,pLine,width);
			for(int x=0;x<=(right-left);x++){
				if(y == 0){ //@ŏ
				//	pShift[x] = (pShift[x] >> 8) & ~topMusk;
					pShift[x] |= ((uint16_t)(pgm_read_byte(&src[y*width+x+pictOfs])) << lShift);
					pLine[x] &= topMusk;
					pLine[x] |= pShift[x] & ~topMusk;
				}
				else if(y == (b-t-1)){ // Ō̍s
					pShift[x] = ((pShift[x]&0xFF00) >> 8) & topMusk;
					pShift[x] |= ((uint16_t)(pgm_read_byte(&src[y*width+x+pictOfs])) << lShift) & ~topMusk;
					if(rShift){
						pLine[x] &= ~endMusk;
						pLine[x] |= pShift[x] & endMusk;
					}else{
						pLine[x] = pShift[x];
					}
				}
				else /**if((y != 0) && (y != (b-t-1)))**/{ // ԍs (O̍s̎ciE)ƍ̍s̍
					pShift[x] = ((pShift[x]&0xFF00) >> 8) & topMusk;
					pShift[x] |= ((uint16_t)(pgm_read_byte(&src[y*width+x+pictOfs])) << lShift) & ~topMusk;
					pLine[x] = pShift[x] & 0xff;
				}
			}
			WriteLcdPageRam(left,y+t,pLine,width);
		}
	}
	
	free(lineBuff);
	free(shiftBuff);
	return 1;
}

// ʃf[^[vOkbcɏށiʒuRECTŎwj
int WriteLcdFromPrgmR(RECT rect,const prog_uchar src[]){
	return WriteLcdFromPrgm(rect.left,rect.top,src);
}


// q`l̉ʃf[^[kbcɏ
// f[^[̓wb_[Ȃ̐摜
// ߂FŔO
int WriteLcdFromRam(int left,int top
	,int right,int bottom
	,const uint8_t src[]){
	
	int width = right-left+1;
	int t = top/8;
	int b = bottom/8 + 1; 


	if((!top%8)&&(!(bottom+1)%8)){ // ㉺ƂɃoCgEȂ̂ŒP`̂
		WritePageRectRam(left,t,right,b,src);
		return 1; // Ń^[
	}
	int lShift = top%8; // oCgÊ
	// sڂ̃}XN
	uint8_t topMusk = 0;
	if(lShift){
		for(int i=0;i<lShift;i++){
			topMusk = topMusk << 1;
			topMusk |= 1;
		}
	}// topMusk ́@obN}XN

	// Ō̍s̃}XN

	int rShift = (bottom+1)%8;
	uint8_t endMusk = 0;
	if(rShift){
		for(int i=0;i<rShift;i++){
			endMusk = endMusk << 1;
			endMusk |= 1;
		}
	}// endMusk̓f[^[}XN
	
	uint8_t *pLine = malloc(width);
	uint16_t *pShift = malloc(width<<1);
	if(pShift==NULL){
		DisplayOn(0);
		return 0;
	}
	if((b-t)<2){ // s̏̂
		if(rShift) topMusk = topMusk | ~endMusk;
		ReadLcdPage(left,t,pLine,right-left+1);
		for(int x= 0;x<=right-left;x++){
			pLine[x] &= topMusk;
			pLine[x] |= ((src[x] << lShift) & ~topMusk);
		}
		WriteLcdPageRam(left,t,pLine,right-left+1);
	}else{	// s
		for(int i=0;i<(right-left+1);i++){// Vtgobt@NA
			pShift[i]=0;
		}
		for(int y=0;y<b-t;y++){
			ReadLcdPage(left,y+t,pLine,(right-left+1));
			for(int x=0;x<=(right-left);x++){
				if(y == 0){ //@ŏ
					pShift[x] |= ((uint16_t)src[y*width+x] << lShift);
					pLine[x] &= topMusk;
					pLine[x] |= pShift[x] & ~topMusk;
				}else if(y == (b-t-1)){ // Ō̍s
					pShift[x] = ((pShift[x]&0xFF00) >> 8) & topMusk;
					pShift[x] |= ((uint16_t)src[y*width+x] << lShift) & ~topMusk;
					if(rShift){
						pLine[x] &= ~endMusk;
						pLine[x] |= pShift[x] & endMusk;
					}else{
						pLine[x] = pShift[x];
					}
				}else{ // ԍs (O̍s̎ciE)ƍ̍s̍
					pShift[x] = ((pShift[x]&0xFF00) >> 8) & topMusk;
					pShift[x] |= ((uint16_t)src[y*width+x] << lShift) & ~topMusk;
					pLine[x] = pShift[x] & 0xff;
				}
			}
			WriteLcdPageRam(left,y+t,pLine,right-left+1);
		}
	}
	
	free(pLine);
	free(pShift);
	return 1;
}

// q`l̉ʃf[^[kbcɏށiʒuRECTŎwj
int WriteLcdRamR(RECT rect,const uint8_t src[]){
	return WriteLcdFromRam(rect.left,rect.top,rect.right,rect.bottom,src);
}

// Horizontal Bar Graph  Value Range [0..1000]
// o[Ot@W͂OPOOOtXP[Œ
// 
// positionF	ʒu
// LHomeF		Oō_AO͉E_
// padShapeF	OŃnb`OAOŃXvbĝ
// ߂F		ŔO
int BarGraphH(int left,int top,int right,int bottom,
	int position,uint8_t LHome,uint8_t padShape){
	int width = abs(right-left+1);
	int height = abs(bottom-top+1);
	uint8_t *buff = malloc(width * ((height>>3)+1));
	if(buff == NULL){
		DisplayOn(0);
		return 0;
	}
	for(int i=0;i<(width*((height>>3)+1));i++){
		buff[i]=0;
	}
	if(position<0)position = 0;
	if(position>1000)position = 1000;
	int pos = (int)((long)width*(long)position/(long)1000);//\ʒuZo
	int xp;
	for(int x=0;x<width;x++){
		for(int y=0;y<height;y++){
			if(LHome)xp=x;else xp=width-1-x;
			if(xp==pos){
				buff[(y/8)*width+x] |= (1<<(y%8));
			}else if((padShape)&&(xp<pos)){
				buff[(y/8)*width+x] |= (1<<(y%8));
			}
		}
	}
	WriteLcdFromRam(left,top,right,bottom,buff);				
	free(buff);
	return 1;
}

// o[OtiʒuRECTŎwj
int BarGraphHR(RECT rect,int position,uint8_t LHome,uint8_t padShape){
	return BarGraphH(rect.left,rect.top,rect.right,rect.bottom,position,LHome,padShape);
}

// Vertical Bar Graph [0..1000]
// o[Ot
// W@OPOOOtXP[
// positionF	ʒuixj
// DHomeF		Oŉ_AOŏ㌴_
// padSahape:	OŃnb`OAOŃXvbĝ
// ߂F		ŔO
int BarGraphV(int left,int top,int right,int bottom,
	int position, uint8_t DHome,uint8_t padShape){
	int width = abs(right-left+1);
	int height = abs(bottom-top+1);
	uint8_t *buff = malloc(width * ((height>>3)+1));
	if(buff == NULL){
		DisplayOn(0);
		return 0;
	}
	for(int i=0;i<(width*((height>>3)+1));i++){
		buff[i]=0;
	}
	if(position<0)position = 0;
	if(position>1000) position = 1000;
	int pos = (int)((long)height*(long)position/(long)1000);
	int yp;
	for(int y=0;y<height;y++){
		for(int x=0;x<width;x++){
			if(DHome)yp=height-1-y;else yp = y;
			if(yp==pos){
				buff[(y/8)*width+x] |= (1<<(y%8));
			}else if((padShape)&&(yp<pos)){
				buff[(y/8)*width+x] |= (1<<(y%8));
			}
		}
	}
	WriteLcdFromRam(left,top,right,bottom,buff);
	free(buff);
	return 1;
}

//o[Ot\iʒuRECTŎwj
int BarGraphVR(RECT rect,int position, uint8_t DHome,uint8_t padShape){
	return BarGraphV(rect.left,rect.top,rect.right,rect.bottom,position,DHome,padShape);
}


// Display String
// \
// 
// str:	
// font: tHg
// ߂F	OŐ
int PutStr(int left,int top,int right,int bottom
	,char *str,const prog_uchar font[]){

	int width = right-left+1;
	int height = bottom-top+1;
	int lines = height>>3; // /8;
	if(height%8) lines++;
	int fontOffs = ((pgm_read_byte(&font[0])<<8) & 0xFF00);
	fontOffs |= ((pgm_read_byte(&font[1])) & 0xFF);
	int fontX = (pgm_read_byte(&font[2])<<8) & 0xFF00;
	fontX |= pgm_read_byte(&font[3]) & 0xFF;
	int fontY = (pgm_read_byte(&font[4])<<8) &0xFF00;
	fontY |= pgm_read_byte(&font[5]) & 0xFF;
	int fontLines = fontY >> 3; // /8;
	if(fontY%8) fontLines ++;


	uint8_t *buff = malloc(width*lines);
	for(int i=0;i<(width*lines);i++){
		buff[i] = 0;
	}
	int xPos = 0;
	int fontN;
	while(*str){
		fontN = *str-fontOffs;
		for(int x=0;x<fontX;x++){
			for(int i=0;i<lines;i++){
				if(i<=fontLines)buff[(i*width)+xPos]
					=pgm_read_byte(&font[FONT_OFFSET+(fontN*fontLines*fontX)+(x*fontLines)+i]); // Y
			}
			xPos++;
		}
		if(xPos>=width)break;
		str++;
	}
	WriteLcdFromRam(left,top,right,bottom,buff);
	free(buff);
	return 1;
}

// \@iʒuRECTŎwj
int PutStrR(RECT rect,char *str,const prog_uchar font[]){
	return PutStr(rect.left,rect.top,rect.right,rect.bottom,str,font);
}


// Display String Vertical
// \@E菑iցj
// 
// str:	
// font:	tHg
// ߂F	OŐ
int PutStrV(int left,int top,int right, int bottom
	,char *str,const prog_uchar font[]){

	int width = right-left+1;
	int height = bottom-top+1;
	int lines = height>>3; // /8;
	if(height%8) lines++;
	int fontOffs = ((pgm_read_byte(&font[0])<<8) & 0xFF00);
	fontOffs |= ((pgm_read_byte(&font[1])) & 0XFF);
	int fontX = (pgm_read_byte(&font[2])<<8) & 0xFF00;
	fontX |= pgm_read_byte(&font[3]) & 0XFF;
	int fontY = (pgm_read_byte(&font[4])<<8) & 0xFF00;
	fontY |= pgm_read_byte(&font[5]) & 0xFF;
	int fontLines = fontY >> 3; // /8;
	if(fontY%8) fontLines ++;


	uint8_t *buff = malloc(width*lines);
	if(buff == NULL){
		DisplayOn(0);
		for(;;);
	}
	for(int i=0;i<(width*lines);i++){
		buff[i] = 0;
	}

	int yPos = height-1;
	int fontN;
	uint8_t fontLine[fontLines];

	while(*str){
		fontN = *str - fontOffs;
		str++;	
		for(int x=0;x<fontX;x++){ // P̃C疄߂
			for(int i=0;i<fontLines;i++){// tHġ@C擾
				fontLine[i] =  pgm_read_byte(&font[FONT_OFFSET+(fontN*fontLines*fontX)+(x*fontLines)+i]);
			}
			// 1Cobt@ɃrbgɓWJB
			int l = width-1; // Lrbg
			for(int i=0;i<fontLines;i++){
				for(int n=0;n<8;n++){ // 
					if(l>=0) // LȋЂ͈̔͂
						if(fontLine[i] & (1<<n)) buff[(yPos>>3)*width + n +(i<<3)] |= (1 << (yPos%8));
					l--;
				}
			} 
			yPos--;
			if(yPos<0)goto LoopEnd;
		}
	}
LoopEnd:
	
	WriteLcdFromRam(left,top,right,bottom,buff);
	free(buff);
	return 1;
}

// \iʒuRECTŎwj
int PutStrVR(RECT rect,char *str,const prog_uchar font[]){
	return PutStrV(rect.left,rect.top,rect.right,rect.bottom,str,font);
}


